### client.perl  -*- Perl -*-
## An interactive client for the logical testing interface protocol.

### Ivan Shmakov, 2022

## To the extent possible under law, the author(s) have dedicated
## all copyright and related and neighboring rights to this software
## to the public domain worldwide.  This software is distributed
## without any warranty.

## You should have received a copy of the CC0 Public Domain Dedication
## along with this software.  If not, see
## <http://creativecommons.org/publicdomain/zero/1.0/>.

### History:

## 0.3  2022-10-11 17:35Z
##      Fixed: provide an anonymous function as the first argument
##      to Lotei::Client::with_timeout explicitly (was: trying to
##      rely on the prototype.)

## 0.2  2022-10-08 16:37:26Z
##      (sfn.iAwLvcfyt_m6XL6kV5W31UHNz-OseSNbVu3Q4SNVPZQ.perl)
##      (Lotei::Client): Require module.  Update the main loop
##      to use the functions provided there.
##      (fletcher): Removed function, now in Lotei::Client.
##      (read_up): Likewise.
##      (read_response): Likewise.
##      (xdump): Likewise.

## 0.1  2022-10-05 10:35:28Z
##      (sfn.RtVr6-6f6hrlUy-qbrUjBxa-Q9X4p3xvQlbaBXxiHW8.perl)
##      Initial revision.

### Code:
use common::sense;
use English qw (-no_match_vars);

require IO::Handle;
require IO::Socket::UNIX;
require Lotei::Client;

our $usage = qq {Usage:
    \$ ${0} [--debug] [--] TTY-OR-SOCKET 
};

our ($prefix)
    = ("");

## GNU error function look-alike (for possible future i18n)
sub gerr
{
    my $exit = shift ();
    my $s = &CORE::sprintf (@_);
    $s .= "\n"
        unless ("\n" eq substr ($s, -1));
    print STDERR ($s);

    ## .
    exit ($exit)
        unless (0 == $exit);
}

## FIXME: implement proper arguments parsing
our ($debug_p)
    = 0;
if ($ARGV[0] eq substr ("--debug", 0, length ($ARGV[0]))) {
    $debug_p
        = 1;
    shift (@ARGV);
}
if ("--" eq $ARGV[0]) {
    shift (@ARGV);
}
## .
gerr (1, "%s", $usage)
    unless (1 == @ARGV);
our $endp
    = shift (@ARGV);

my $eh;
if (-S $endp) {
    ## .
    $eh = IO::Socket::UNIX->new  ("Type", IO::Socket::SOCK_STREAM (),
                                  "Peer", $endp)
        or  gerr (1, "%s: Cannot connect to socket: %s",
                  $endp, $IO::Socket::errstr);
} else {
    ## .
    open ($eh, "+<", $endp)
        or gerr (1, "%s: Cannot open: %s", $endp, $!);
}
# $eh->binmode ();
$eh->autoflush (1);

while (<STDIN>) {
    chomp ();
    unless (m {^(?: \s*
                      (?<debug> debug)
                    | (?<x> [[:blank:]0-9a-f]+)
                    ) $
            }xi) {
        gerr (0, "Input line not understood; ignored");
        next;
    }

    if (exists ($LAST_PAREN_MATCH{"debug"})) {
        $debug_p
            = ! $debug_p;
        warn ("D: Debug mode on\n")
            if ($debug_p);
        next;
    }

    s/\s+//g;
    next
        unless (length () > 0);
 
    my $s
        = Lotei::Client::frame (pack ("H*", $_));
    warn ("D: Sending: ", Lotei::Client::xdump ($s), "\n")
        if ($debug_p);
    $eh->print ($s);

    my $r = eval {
        ## FIXME: hardcoded timeout
        ## .
        Lotei::Client::with_timeout (sub {
            ## .
            Lotei::Client::read_response ($eh);
        }, 7);
    };
    if (! $@) {
        ## do nothing
    } elsif ("*timeout*\n" eq $@) {
        gerr (0, "Timeout reading response from remote");
        next;
    } else {
        ## .
        gerr (1, "Error reading response from remote: %s", $@ // $!);
    }
    print ("Got response: ", Lotei::Client::xdump ($r), "\n");
}

### client.perl ends here
