/*** lotei.c  -*- C -*- */
/** Primary module, logical testing interface. */

/*** Copyright (C) 2022 Ivan Shmakov */

/** Permission to copy this software, to modify it, to redistribute it,
 ** to distribute modified versions, and to use it for any purpose is
 ** granted, subject to the following restrictions and understandings.
 **
 ** 1.  Any copy made of this software must include this copyright notice
 **     in full.
 **
 ** 2.  I have made no warranty or representation that the operation of
 **     this software will be error-free, and I am under no obligation to
 **     provide any services, by way of maintenance, update, or otherwise.
 **
 ** 3.  In conjunction with products arising from the use of this material,
 **     there shall be no use of my name in any advertising, promotional,
 **     or sales literature without prior written consent in each case.
 */

/*** History: */

/** 0.4 2022-10-07 21:10Z
 **     (LOTEI_PORT_OUT_MASK_1): New user-overridable macro replacing
 **     (LOTEI_PORT_OUT_MASK): this now computed one.
 **     (LOTEI_PORT_IN_MASK_1): New user-overridable macro replacing
 **     (LOTEI_PORT_IN_MASK): this now computed one.
 **     (bitcount): New unit-local function.
 **     (check_0): Likewise.
 **     (main): Implement the device interaction request.  Call
 **     readyled_blink whenever a response is prepared (was: only
 **     on successful responses.)
 **
 ** 0.3 2022-10-07 17:10:26Z
 **     (sfn.WuLImzfmKRuC0zc-iDnefSrG53IQoL79h9vVduFUYPg.c)
 **     (stuff_head): New unit-local function split off
 **     (stuff_ack): this
 **     (stuff_err): and this functions.
 **     (main): Implement the response retrieval request.  Assert CTS
 **     just before checking the UART readiness (was: when there are
 **     no more octets to write in the response buffer.)
 **
 ** 0.2 2022-10-07 09:55:23Z
 **     (sfn.euEIhiJg5gF45mT6fU5BBG1-q2DjcPeA61rc3uJwVCk.c)
 **     (add_255): New unit-local inline function split off
 **     (stuff): this function.
 **     (main): Use add_255 (was: 16-bit arithmetic and % operator.)
 **     Fixed: check for 0x88 as the last octet of the opaque protocol
 **     version identifier (was: 88.)  Set seen_handshake_p to zero in
 **     case of an unrecognized version identifier in the handshake
 **     request (was: not changed.)
 **
 ** 0.1 2022-10-07 08:47:41Z
 **     (sfn.e7nGYxKlfrEcDsAXCi-lSGXxJ-HqCZVh37eOqviQZy8.c)
 **     Initial revision.
 */

/*** Code: */
#include <stdbool.h>
#include <stdint.h>

#include <machine.h>
#include <readyled.h>

#ifndef MACHINE_MEGA8_LIKE
#define MACHINE_MEGA8_LIKE  (0)
#endif
#ifndef MACHINE_MEGA64_LIKE
#define MACHINE_MEGA64_LIKE (0)
#endif

#ifndef LOTEI_PORT_CTS_INVERSE
#define LOTEI_PORT_CTS_INVERSE  (0)
#endif

#if ! defined (LOTEI_PORT_CTS_SHIFT) && ! defined (LOTEI_PORT_CTS_MASK)
#warning "Neither LOTEI_PORT_CTS_SHIFT nor _MASK defined; NO CTS support"
#define LOTEI_PORT_CTS_MASK     (0ULL)
#define LOTEI_PORT_CTS_ON       (0ULL)
#define LOTEI_PORT_CTS_OFF      (0ULL)
#else
#ifndef LOTEI_PORT_CTS_MASK
#define LOTEI_PORT_CTS_MASK     (1ULL << (LOTEI_PORT_CTS_SHIFT))
#endif
#ifndef LOTEI_PORT_CTS_ON
#define LOTEI_PORT_CTS_ON \
    ((LOTEI_PORT_CTS_INVERSE) ? 0 : (LOTEI_PORT_CTS_MASK))
#endif
#ifndef LOTEI_PORT_CTS_OFF
#define LOTEI_PORT_CTS_OFF \
    ((LOTEI_PORT_CTS_ON)          ^ (LOTEI_PORT_CTS_MASK))
#endif
#endif

#ifdef LOTEI_PORT_IN_SHIFT
/* do nothing */
#elif MACHINE_MEGA8_LIKE
/** Use PORTB for output and PORTC 0.5 for input on ATmega8 and the like */
#warning "FIXME: check MACHINE_MEGA8_LIKE support"
#define LOTEI_PORT_IN_SHIFT     (16)
#define LOTEI_PORT_IN_MASK_1    (0077ULL)
#define LOTEI_PORT_OUT_SHIFT     (8)
#define LOTEI_PORT_OUT_MASK_1   (0377ULL)
#elif MACHINE_MEGA64_LIKE
/** Use PORTC for output and PORTA for input on ATmega64 and the like */
#define LOTEI_PORT_IN_SHIFT      (0)
#define LOTEI_PORT_IN_MASK_1    (0377ULL)
#define LOTEI_PORT_OUT_SHIFT    (16)
#define LOTEI_PORT_OUT_MASK_1   (0377ULL)
#else
#error "FIXME: unknown MCU? define LOTEI_PORT_ macros explicitly?"
#endif

#define LOTEI_PORT_IN_MASK \
    ((LOTEI_PORT_IN_MASK_1) << (LOTEI_PORT_IN_SHIFT))
#define LOTEI_PORT_OUT_MASK \
    ((LOTEI_PORT_OUT_MASK_1) << (LOTEI_PORT_OUT_SHIFT))

#ifndef LOTEI_TIMER_DIVISOR
#define LOTEI_TIMER_DIVISOR (1024L)
#endif

enum {
  REQ_HANDSHAKE = 002,
  REQ_IF_CONF   = 004,
  REQ_INTERACT  = 020,
  REQ_RETRIEVE  = 022
};

enum {
  RESP_ACK      = 001,
  RESP_ERR      = 003,
  RESP_RESULT   = 023
};

enum {
  ERR_TYPE_CODE = 001,
  ERR_DATA_LEN  = 002,
  ERR_NOT_SUPP  = 003,
  ERR_EXCEED    = 004,
  ERR_IMPL_RES  = 005
};

static inline uint8_t
add_255 (uint8_t a, uint8_t b)
{
  uint8_t r
    = a + b;
  /* . */
  return (r == 255 ? 0 : r + (r < a ? 1 : 0));
}

static inline uint_least8_t
bitcount (uint8_t v)
{
  uint_least8_t r;
  for (r = 0; v > 0; v >>= 1) {
    if ((1 & v)) { ++r; }
  }
  /* . */
  return r;
}

static inline bool
check_0 (uint8_t *p, uint_least8_t z)
{
  for (; z > 0 && *(p++) == '\0'; z--) {
    /* do nothing */
  }
  /* . */
  return (z == 0);
}

static inline void
stuff (uint8_t *sum, uint8_t **bufp, uint8_t val)
{
  sum[1]
    = add_255 (sum[1], (*((*bufp)++) = val));
  sum[0]
    = add_255 (sum[0], sum[1]);
}

static inline void
stuff_head (uint8_t *sum, uint8_t **bufp, uint8_t code)
{
  sum[0] = sum[1] = 0;
  stuff (sum, bufp, code);
  stuff (sum, bufp, (sum - (1 + *bufp)));
}

static inline void
stuff_ack (uint8_t **bufp)
{
  stuff_head (2 + *bufp, bufp, RESP_ACK);
}

static inline void
stuff_err (uint8_t **bufp, uint8_t err)
{
  uint8_t *sp = 3 + *bufp;
  stuff_head (sp, bufp, RESP_ERR);
  stuff (sp, bufp, err);
}

int
main ()
{
  machine_init ();
  machine_timer_setup (LOTEI_TIMER_DIVISOR);
  machine_uart_setup (0);

  machine_port_setup ((  LOTEI_PORT_IN_MASK),
                      (  LOTEI_PORT_OUT_MASK
                       | LOTEI_PORT_CTS_MASK
                       | READYLED_PORT_MASK));

  /** Reading requests */
  bool seen_handshake_p = 0, checksum_ok_p;
  uint_least8_t rd_state = 0, rd_r = 0, rd_i;
  uint8_t rd_code, rd_d[256];
  uint8_t rd_s0, rd_s1;

  /** Writing responses */
  uint_least16_t wd_r = 0;
  uint8_t wd_d[260], *wd_p;

  /** Interacting with the device under test */
  uint8_t in_d[260];
  uint8_t *in_p, *in_sp;
  uint_least8_t in_r = 0;
  uint_least8_t in_i, in_gm, in_sm;

#if 1
  /** NB: mainly to suppress warnings */
  checksum_ok_p
    = 0;
  rd_s1
    = rd_s0
    = 0;
#endif

  while (1) {
    machine_wait ();

    if (machine_timer_ticked_p ()) {
      readyled_tick ();
    }

    /** Write an octet of response, if any */
    if (wd_r <= 0) {
      /** No response in the buffer; do nothing */
    } else if (machine_uart_can_write_p (0)) {
      --wd_r;
      machine_uart_write (0, *(wd_p++));
    } else {
      /** Wait until UART becomes available */
      continue;
    }

    /** Interact with the device once, if requested */
    if (in_r > 0) {
      if (rd_d[in_i]-- > 0) {
        /** Get data from device, according to in_gm */
        uint_least8_t gm = in_gm;
        if ((1 & gm)) {
          machine_port_map_t v
            = machine_port_read  (LOTEI_PORT_IN_SHIFT,
                                  LOTEI_PORT_IN_MASK_1);
          stuff (in_sp, &in_p, (uint8_t)v);
        }
        /* gm >>= 1; if ((1 & gm)) { } */
        continue;
      }

      /** Send data to device, according to in_sm */
      --in_r, ++in_i;
      uint_least8_t sm = in_sm;
      if ((1 & sm)) {
        machine_port_map_t v
          = rd_d[in_i++];
        machine_port_write (((v & LOTEI_PORT_OUT_MASK_1)
                             << LOTEI_PORT_OUT_SHIFT),
                            (LOTEI_PORT_OUT_MASK));
      }
      /* sm >>= 1; if ((1 & sm)) { } */
    }

    /** Process a fully received frame, if any */
    if (4 == rd_state) {
      rd_state
        = 0;
      wd_p
        = wd_d;
      if (! checksum_ok_p) {
        /** Checksum wrong; ignore */
        /* do nothing */
      } else if (! seen_handshake_p && rd_code != REQ_HANDSHAKE) {
        /** Command before handshake */
        stuff_err (&wd_p, ERR_TYPE_CODE);
      } else if (rd_code == REQ_HANDSHAKE) {
        /** Handshake received */
        if (4 != rd_i || rd_d[0] != 0x24 || rd_d[1] != 0x3f
            || rd_d[2] != 0x6a || rd_d[3] != 0x88) {
          /** Unrecognized (unsupported) protocol */
          seen_handshake_p = 0;
          stuff_err (&wd_p, ERR_NOT_SUPP);
        } else {
          seen_handshake_p = 1;
          stuff_ack (&wd_p);

          /** Place an empty response into in_d */
          in_p = in_d;
          in_sp = 2 + in_d;
          stuff_head (in_sp, &in_p, RESP_RESULT);
        }
      } else if (rd_code == REQ_INTERACT) {
        uint8_t *stop;
        uint8_t *tp, *cp;
        uint8_t r_st;
        if (2 > rd_i
            || (tp = rd_d + 1 + *rd_d)  > (stop = rd_d + rd_i)
            || (cp = tp   + 1 + *tp)    >  stop) {
          /** The bitmaps do not seem to fit in the request */
          stuff_err (&wd_p, ERR_DATA_LEN);
        } else if  ((*rd_d  > 1 && ! check_0 (1 + rd_d, -1 + *rd_d))
                    || (*tp > 1 && ! check_0 (1 + tp,   -1 + *tp))
                    || (*rd_d == 0 ? (in_gm = 0)
                        : 0 != ((~ 1) & (in_gm = *(rd_d + *rd_d))))
                    || (*tp   == 0 ? (in_sm = 0)
                        : 0 != ((~ 1) & (in_sm = *(tp   + *tp))))) {
          /** Reserved (unsupported) bits are non-zero */
          stuff_err (&wd_p, ERR_NOT_SUPP);
        } else if (0 != (stop - cp) % (r_st = 1 + bitcount (in_sm))) {
          /** Not an integral number of transfer instructions */
          stuff_err (&wd_p, ERR_DATA_LEN);
        } else {
          uint_least16_t w;
          uint_least16_t w_st = bitcount (in_gm);
          in_i = cp - rd_d;
          for  (in_r = 0, w = 0; cp < stop;
                w += w_st * (*cp), cp += r_st, in_r++) {
            if (w > 255) {
              break;
            }
          }
          if (w > 255) {
            /** The response will not fit a single frame */
            in_r = 0;
            stuff_err (&wd_p, ERR_EXCEED);
          } else {
            /** Place a response header into in_d */
            in_p = in_d;
            in_sp = 2 + w + in_d;
            stuff_head (in_sp, &in_p, RESP_RESULT);
            /** And also the immediate acknowledgement into wd_d */
            stuff_ack (&wd_p);
          }
        }
      } else if (rd_code == REQ_RETRIEVE) {
        /** New response from alternative buffer */
        readyled_blink ();
        wd_r = 2 + in_p - in_d;
        wd_p = in_d;
        continue;
      } else {
        /** Unrecognized command code */
        stuff_err (&wd_p, ERR_TYPE_CODE);
      }

      if (wd_p != wd_d) {
        /** New response was produced; recompute and skip to the beginning */
        readyled_blink ();
        wd_r = 2 + wd_p - wd_d;
        wd_p = wd_d;
        continue;
      }
    }

    /** We are now ready to receive data; assert CTS */
    machine_port_write ((LOTEI_PORT_CTS_ON),
                        (LOTEI_PORT_CTS_MASK));

    /** Read a request frame octet */
    if (machine_uart_can_read_p (0)) {
      uint_least8_t u
        = machine_uart_read (0);
      /** Deassert CTS */
      machine_port_write   ((LOTEI_PORT_CTS_OFF),
                            (LOTEI_PORT_CTS_MASK));
      if (rd_r > 0) {
        --rd_r;
        rd_d[rd_i++]
          = u;
      } else if (1 == ++rd_state) {
        rd_s1 = rd_s0 = 0;
        rd_code
          = u;
      } else if (2 == rd_state) {
        rd_i = 0;
        rd_r = u;
      } else if (3 == rd_state) {
        checksum_ok_p
          = (rd_s1 == u);
        continue;
      } else {
        checksum_ok_p
          = checksum_ok_p && (rd_s0 == u);
        continue;
      }
      rd_s0
        = add_255 (rd_s0, u);
      rd_s1
        = add_255 (rd_s1, rd_s0);
    }
  }

  /* not reached */
}

/*** lotei.c ends here */
